@@ -4,5 +4,5 @@ from .TD2 import TD2, TD2_Reynolds_Correction
 from .ainleymathieson import AinleyMathieson
 from .fixedefficiency import FixedEfficiency
 from .fixedpressureloss import FixedPressureLoss
-from .kackerokapuu import KrackerOkapuu
+from .kackerokapuu import KackerOkapuu
 from .traupel import Traupel 
\ No newline at end of file
